/*
 * Copyright (C) 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "SharedMemory.h"
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WTF {
    class String;
}

namespace WebCore {
    struct PasteboardWebContent;
    struct PasteboardImage;
    class Color;
}

namespace WebKit {
    struct WebSelectionData;
}

namespace Messages {
namespace WebPasteboardProxy {

static inline IPC::StringReference messageReceiverName()
{
    return IPC::StringReference("WebPasteboardProxy");
}

#if PLATFORM(IOS)
class WriteWebContentToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardWebContent&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteWebContentToPasteboard"); }
    static const bool isSync = false;

    WriteWebContentToPasteboard(const WebCore::PasteboardWebContent& content, const String& pasteboardName)
        : m_arguments(content, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class WriteImageToPasteboard {
public:
    typedef std::tuple<const WebCore::PasteboardImage&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteImageToPasteboard"); }
    static const bool isSync = false;

    WriteImageToPasteboard(const WebCore::PasteboardImage& pasteboardImage, const String& pasteboardName)
        : m_arguments(pasteboardImage, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class WriteStringToPasteboard {
public:
    typedef std::tuple<const String&, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteStringToPasteboard"); }
    static const bool isSync = false;

    WriteStringToPasteboard(const String& pasteboardType, const String& text, const String& pasteboardName)
        : m_arguments(pasteboardType, text, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class ReadStringFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadStringFromPasteboard"); }
    static const bool isSync = true;

    typedef std::tuple<String&> Reply;
    ReadStringFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class ReadURLFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadURLFromPasteboard"); }
    static const bool isSync = true;

    typedef std::tuple<String&> Reply;
    ReadURLFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class ReadBufferFromPasteboard {
public:
    typedef std::tuple<uint64_t, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadBufferFromPasteboard"); }
    static const bool isSync = true;

    typedef std::tuple<WebKit::SharedMemory::Handle&, uint64_t&> Reply;
    ReadBufferFromPasteboard(uint64_t index, const String& pasteboardType, const String& pasteboardName)
        : m_arguments(index, pasteboardType, pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS)
class GetPasteboardItemsCount {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardItemsCount"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    explicit GetPasteboardItemsCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardTypes {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardTypes"); }
    static const bool isSync = true;

    typedef std::tuple<Vector<String>&> Reply;
    explicit GetPasteboardTypes(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardPathnamesForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardPathnamesForType"); }
    static const bool isSync = true;

    typedef std::tuple<Vector<String>&> Reply;
    GetPasteboardPathnamesForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardStringForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardStringForType"); }
    static const bool isSync = true;

    typedef std::tuple<String&> Reply;
    GetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardBufferForType {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardBufferForType"); }
    static const bool isSync = true;

    typedef std::tuple<WebKit::SharedMemory::Handle&, uint64_t&> Reply;
    GetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType)
        : m_arguments(pasteboardName, pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class PasteboardCopy {
public:
    typedef std::tuple<const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("PasteboardCopy"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    PasteboardCopy(const String& fromPasteboard, const String& toPasteboard)
        : m_arguments(fromPasteboard, toPasteboard)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardChangeCount {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardChangeCount"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    explicit GetPasteboardChangeCount(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardUniqueName {
public:
    typedef std::tuple<> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardUniqueName"); }
    static const bool isSync = true;

    typedef std::tuple<String&> Reply;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardColor {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardColor"); }
    static const bool isSync = true;

    typedef std::tuple<WebCore::Color&> Reply;
    explicit GetPasteboardColor(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetPasteboardURL {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("GetPasteboardURL"); }
    static const bool isSync = true;

    typedef std::tuple<String&> Reply;
    explicit GetPasteboardURL(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AddPasteboardTypes {
public:
    typedef std::tuple<const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("AddPasteboardTypes"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    AddPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardTypes {
public:
    typedef std::tuple<const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardTypes"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    SetPasteboardTypes(const String& pasteboardName, const Vector<String>& pasteboardTypes)
        : m_arguments(pasteboardName, pasteboardTypes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardPathnamesForType {
public:
    typedef std::tuple<const String&, const String&, const Vector<String>&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardPathnamesForType"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    SetPasteboardPathnamesForType(const String& pasteboardName, const String& pasteboardType, const Vector<String>& pathnames)
        : m_arguments(pasteboardName, pasteboardType, pathnames)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardStringForType {
public:
    typedef std::tuple<const String&, const String&, const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardStringForType"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    SetPasteboardStringForType(const String& pasteboardName, const String& pasteboardType, const String& string)
        : m_arguments(pasteboardName, pasteboardType, string)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetPasteboardBufferForType {
public:
    typedef std::tuple<const String&, const String&, const WebKit::SharedMemory::Handle&, uint64_t> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("SetPasteboardBufferForType"); }
    static const bool isSync = true;

    typedef std::tuple<uint64_t&> Reply;
    SetPasteboardBufferForType(const String& pasteboardName, const String& pasteboardType, const WebKit::SharedMemory::Handle& handle, uint64_t size)
        : m_arguments(pasteboardName, pasteboardType, handle, size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class WriteToClipboard {
public:
    typedef std::tuple<const String&, const WebKit::WebSelectionData&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("WriteToClipboard"); }
    static const bool isSync = false;

    WriteToClipboard(const String& pasteboardName, const WebKit::WebSelectionData& pasteboardContent)
        : m_arguments(pasteboardName, pasteboardContent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class ReadFromClipboard {
public:
    typedef std::tuple<const String&> Arguments;

    static IPC::StringReference receiverName() { return messageReceiverName(); }
    static IPC::StringReference name() { return IPC::StringReference("ReadFromClipboard"); }
    static const bool isSync = true;

    typedef std::tuple<WebKit::WebSelectionData&> Reply;
    explicit ReadFromClipboard(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebPasteboardProxy
} // namespace Messages
